"""
pyrad.io.read_data_other
========================

Functions for reading auxiliary data

.. autosummary::
    :toctree: generated/

    read_centroids_npz
    read_centroids
    read_proc_periods
    read_profile_ts
    read_histogram_ts
    read_quantiles_ts
    read_vpr_theo_parameters
    read_rhi_profile
    read_last_state
    read_status
    read_mch_xml_vad
    read_rad4alp_icon
    read_rad4alp_vis
    read_mf_vis
    read_histogram
    read_quantiles
    read_excess_gates
    read_colocated_gates
    read_colocated_data
    read_colocated_data_time_avg
    read_timeseries
    read_ts_cum
    read_ml_ts
    read_monitoring_ts
    read_monitoring_ts_old
    read_intercomp_scores_ts
    read_intercomp_scores_ts_old
    read_intercomp_scores_ts_old_v0
    read_selfconsistency
    read_antenna_pattern
    read_xls

"""

import os
import glob
import datetime

import pandas as pd
import csv
import xml.etree.ElementTree as et
from warnings import warn

try:
    import fcntl

    FCNTL_AVAIL = True
except ModuleNotFoundError:
    import msvcrt  # For windows

    FCNTL_AVAIL = False
import time
import errno

import numpy as np

import pyart
from pyart.config import get_fillvalue, get_metadata

from .io_aux import get_fieldname_pyart, _get_datetime
from .flock_utils import lock_file, unlock_file


def read_centroids_npz(fname):
    """
    Reads an npz file containing data used in centroid computation

    Parameters
    ----------
    fname : str
        name of the file to read

    Returns
    -------
    centroids_data : dict
        Dictionary with the data

    """
    try:
        with np.load(fname, allow_pickle=True) as npzfile:
            return npzfile["centroids_data"].item()
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_centroids(fname):
    """
    Reads a file containing the centroids for the semi-supervised
    classification

    Parameters
    ----------
    fname : str
        name of the file to read

    Returns
    -------
    mass_centers : 2D-array
        Matrix with the centroids as (nclasses, nvariables)

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nhydro = sum(1 for row in reader)
            nvars = len(reader.fieldnames) - 1
            mass_centers = np.empty((nhydro, nvars))
            hydro_names = []
            var_names = reader.fieldnames[1:]

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                hydro_names.append(row[reader.fieldnames[0]])
                for j, col in enumerate(var_names):
                    mass_centers[i, j] = float(row[col])

            return mass_centers, hydro_names, var_names
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None


def read_proc_periods(fname):
    """
    Reads a file containing the start and stop times of periods to process

    Parameters
    ----------
    fname : str
        name of the file to read

    Returns
    -------
    starttimes, endtimes : array of datetime objects or None
        The start and end times of the periods to process if the reading has
        been successful

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            startimes = np.empty(nrows, dtype=datetime.datetime)
            endtimes = np.empty(nrows, dtype=datetime.datetime)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                startimes[i] = datetime.datetime.strptime(
                    row["start_time"], "%Y%m%d%H%M%S"
                )
                endtimes[i] = datetime.datetime.strptime(
                    row["end_time"], "%Y%m%d%H%M%S"
                )

            return startimes, endtimes
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None


def read_profile_ts(fname_list, labels, hres=None, label_nr=0, t_res=300.0):
    """
    Reads a colection of profile data file and creates a time series

    Parameters
    ----------
    fname_list : str
        list of files to read
    labels : list of str
        The data labels
    hres : float
        Height resolution
    label_nr : int
        the label nr of the data that will be used in the time series
    t_res : float
        time resolution [s]. If None the time resolution is taken as the
        median

    Returns
    -------
    tbin_edges, hbin_edges, np_ma, data_ma, datetime_arr : tupple
        The read data. None otherwise

    """
    data_ma = []
    np_ma = []
    datetime_arr = np.ma.array([], dtype=datetime.datetime)
    for fname in fname_list:
        datetime_arr = np.append(datetime_arr, _get_datetime(fname, "RAINBOW"))
        height, np_t, vals = read_rhi_profile(fname, labels)
        if hres is None:
            hres = np.mean(height[1:] - height[:-1])
        hbin_edges = np.append(height - hres / 2, height[-1] + hres / 2)
        val = vals[:, label_nr]
        data_ma.append(val)
        np_ma.append(np_t)
    data_ma = np.ma.asarray(data_ma)
    np_ma = np.asarray(np_ma, dtype=int)

    # sort data as a function of time
    ind = np.argsort(datetime_arr)
    datetime_arr = datetime_arr[ind]
    data_ma = data_ma[ind, :]
    np_ma = np_ma[ind, :]

    # put date time array as seconds from start of TRT cell
    dt_s = np.empty(datetime_arr.size, dtype=float)
    for j, dt in enumerate(datetime_arr):
        dt_s[j] = (dt - datetime_arr[0]).total_seconds()

    if t_res is None:
        if dt_s.size > 1:
            t_res = np.median(dt_s[1:] - dt_s[:-1])
        else:
            t_res = 300.0
    tbin_edges = np.append(dt_s - t_res, dt_s[-1])

    return tbin_edges, hbin_edges, np_ma, data_ma, datetime_arr


def read_histogram_ts(fname_list, datatype, t_res=300.0):
    """
    Reads a colection of histogram data file and creates a time series

    Parameters
    ----------
    fname_list : str
        list of files to read
    datatype : str
        The data type (dBZ, ZDR, etc.)
    t_res : float
        time resolution [s]. If None the time resolution is taken as the
        median

    Returns
    -------
    tbin_edges, bin_edges, data_ma, datetime_arr : tupple
        The read data. None otherwise

    """
    data_ma = []
    datetime_arr = np.ma.array([], dtype=datetime.datetime)
    for fname in fname_list:
        datetime_arr = np.append(datetime_arr, _get_datetime(fname, "RAINBOW"))
        hist, bin_edges = read_histogram(fname)
        if datatype == "hydro":
            hist[0] = 0
        elif datatype == "RhoHVc":
            hist = hist[bin_edges[0:-1] > 0.95]
            bin_edges = bin_edges[bin_edges > 0.95]
        npoints = np.sum(hist)
        if npoints > 0:
            data_ma.append(hist / np.sum(hist) * 100.0)
        else:
            data_ma.append(hist)
    data_ma = np.ma.asarray(data_ma)

    # sort data as a function of time
    ind = np.argsort(datetime_arr)
    datetime_arr = datetime_arr[ind]
    data_ma = data_ma[ind, :]

    # put date time array as seconds from start of TRT cell
    dt_s = np.empty(datetime_arr.size, dtype=float)
    for j, dt in enumerate(datetime_arr):
        dt_s[j] = (dt - datetime_arr[0]).total_seconds()

    if t_res is None:
        if dt_s.size > 1:
            t_res = np.median(dt_s[1:] - dt_s[:-1])
        else:
            t_res = 300.0
    tbin_edges = np.append(dt_s - t_res, dt_s[-1])

    return tbin_edges, bin_edges, data_ma, datetime_arr


def read_quantiles_ts(fname_list, step=5.0, qmin=0.0, qmax=100.0, t_res=300.0):
    """
    Reads a colection of quantiles data file and creates a time series

    Parameters
    ----------
    fname_list : str
        list of files to read
    step, qmin, qmax : float
        The minimum, maximum and step quantiles
    t_res : float
        time resolution [s]. If None the time resolution is taken as the
        median

    Returns
    -------
    tbin_edges, qbin_edges, data_ma, datetime_arr : tupple
        The read data. None otherwise

    """
    data_ma = []
    datetime_arr = np.ma.array([], dtype=datetime.datetime)
    qbin_edges = np.arange(
        qmin - step / 2, qmax + step / 2 + step / 2, step=step, dtype=float
    )
    qbin_centers = np.arange(qmin, qmax + step / 2, step=step)
    for fname in fname_list:
        values_aux = np.ma.masked_all(qbin_edges.size - 1, dtype=float)
        datetime_arr = np.append(datetime_arr, _get_datetime(fname, "RAINBOW"))

        quantiles, values = read_quantiles(fname)
        if quantiles is not None:
            for i, qbin_center in enumerate(qbin_centers):
                val_aux = values[quantiles == qbin_center]
                if val_aux.size > 0:
                    values_aux[i] = val_aux
        data_ma.append(values_aux)
    data_ma = np.ma.asarray(data_ma)

    # sort data as a function of time
    ind = np.argsort(datetime_arr)
    datetime_arr = datetime_arr[ind]
    data_ma = data_ma[ind, :]

    # put date time array as seconds from start of TRT cell
    dt_s = np.empty(datetime_arr.size, dtype=float)
    for j, dt in enumerate(datetime_arr):
        dt_s[j] = (dt - datetime_arr[0]).total_seconds()

    if t_res is None:
        if dt_s.size > 1:
            t_res = np.median(dt_s[1:] - dt_s[:-1])
        else:
            t_res = 300.0
    tbin_edges = np.append(dt_s - t_res, dt_s[-1])

    return tbin_edges, qbin_edges, data_ma, datetime_arr


def read_vpr_theo_parameters(fname):
    """
    Reads the parameters defining a theoretical VPR profile from a csv file

    Parameters
    ----------
    fname : str
        file name where to store the data

    Returns
    -------
    vpr_theo_dict : dict
        Dictionary containing the parameters

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            vpr_theo_dict = next(reader)
            for k, v in vpr_theo_dict.items():
                vpr_theo_dict[k] = float(v)
            return vpr_theo_dict

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_rhi_profile(
    fname, labels=["50.0-percentile", "25.0-percentile", "75.0-percentile"]
):
    """
    Reads a monitoring time series contained in a csv file

    Parameters
    ----------
    fname : str
        path of time series file
    labels : list of str
        The data labels

    Returns
    -------
    height, np_t, vals : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            nfields = len(labels)
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)

            height = np.empty(nrows, dtype=float)
            np_t = np.zeros(nrows, dtype=int)
            vals = np.ma.empty((nrows, nfields), dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                height[i] = float(row["Altitude [m MSL]"])
                if "N valid" in row:
                    np_t[i] = int(row["N valid"])
                for j, label in enumerate(labels):
                    vals[i, j] = float(row[label])

            vals = np.ma.masked_values(vals, get_fillvalue())

            return height, np_t, vals
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None


def read_last_state(fname):
    """
    Reads a file containing the date of acquisition of the last volume
    processed

    Parameters
    ----------
    fname : str
        name of the file to read

    Returns
    -------
    last_state : datetime object
        the date

    """
    try:
        with open(fname, "r", newline="") as txtfile:
            line = txtfile.readline()
            txtfile.close()
            if not line:
                warn("File " + fname + " is empty.")
                return None
            try:
                last_state = datetime.datetime.strptime(line, "%Y%m%d%H%M%S")
                return last_state
            except ValueError:
                warn("File " + fname + " does not contain a valid date.")
                return None
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_status(voltime, cfg, ind_rad=0):
    """
    Reads rad4alp xml status file.

    Parameters
    ----------
    voltime : datetime object
        volume scan time

    cfg: dictionary of dictionaries
        configuration info to figure out where the data is
    ind_rad: int
        radar index

    Returns
    -------
    root : root element object
        The information contained in the status file

    """
    if cfg["RadarName"] is None:
        raise ValueError(
            "ERROR: Radar Name not specified in config file. \
            Unable to read status data"
        )

    dayinfo = voltime.strftime("%y%j")
    timeinfo = voltime.strftime("%H%M")
    basename = "ST" + cfg["RadarName"][ind_rad] + dayinfo
    if cfg["path_convention"][ind_rad] == "RT":
        datapath = cfg["datapath"][ind_rad] + "ST" + cfg["RadarName"][ind_rad] + "/"
    else:
        datapath = cfg["datapath"][ind_rad] + dayinfo + "/" + basename + "/"
    filename = glob.glob(datapath + basename + timeinfo + "*.xml")
    if not filename:
        warn(
            "rad4alp status file " + datapath + basename + timeinfo + "*.xml not found"
        )
        return None
    try:
        root = et.parse(filename[0]).getroot()
        return root
    except et.ParseError as ee:
        warn(str(ee))
        warn("Unable to read file " + filename[0])

        return None


def read_mch_xml_vad(voltime, cfg, ind_rad=0):
    """
    Reads a MeteoSwiss VAD file (VAx xml files)

    Parameters
    ----------
    voltime : datetime object
        volume scan time

    cfg: dictionary of dictionaries
        configuration info to figure out where the data is
    ind_rad: int
        radar index

    Returns
    -------
    pyart.core.HorizontalWindProfile
        pyart Wind profile object containing the VAD data

    """

    if cfg["RadarName"] is None:
        raise ValueError(
            "ERROR: Radar Name not specified in config file. \
            Unable to read status data"
        )

    dayinfo = voltime.strftime("%y%j")
    timeinfo = voltime.strftime("%H%M")
    basename = "VA" + cfg["RadarName"][ind_rad] + dayinfo
    if cfg["path_convention"][ind_rad] == "RT":
        datapath = cfg["datapath"][ind_rad] + "VA" + cfg["RadarName"][ind_rad] + "/"
    else:
        datapath = cfg["datapath"][ind_rad] + dayinfo + "/" + basename + "/"
    filename = glob.glob(datapath + basename + timeinfo + "*.xml")

    return pyart.aux_io.read_mch_vad(filename)


def read_rad4alp_icon(fname, datatype, ngates=0):
    """
    Reads rad4alp ICON data binary file.

    Parameters
    ----------
    fname : str
        name of the file to read
    datatype : str
        name of the data type
    ngates : int
        maximum number of range gates per ray. If larger than 0 the radar
        field will be cut accordingly.

    Returns
    -------
    field : dictionary
        The data field

    """
    try:
        bindata = np.fromfile(fname, dtype="uint8", count=-1)
        nbins = bindata.size
        naz = 360
        nr = int(nbins / naz)
        bindata = np.reshape(bindata, (naz, nr))
        mask = bindata == 0

        if datatype == "TEMP":
            field_name = get_fieldname_pyart(datatype)
            field_data = np.ma.masked_where(
                mask, (bindata - 1).astype(float) * 0.5 - 87.0
            )

            field = get_metadata(field_name)
            field["data"] = field_data
            if ngates > 0:
                field["data"] = field["data"][:, :ngates]
            return field
        if datatype == "ISO0":
            field_name = get_fieldname_pyart(datatype)
            field_data = np.ma.masked_where(mask, (bindata - 1).astype(float))

            field = get_metadata(field_name)
            field["data"] = field_data
            if ngates > 0:
                field["data"] = field["data"][:, :ngates]
            return field

        warn("Unknown COSMO data type " + datatype)
        return None
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_rad4alp_vis(fname, datatype):
    """
    Reads rad4alp visibility data binary file.

    Parameters
    ----------
    fname : str
        name of the file to read

    datatype : str
        name of the data type

    Returns
    -------
    field_list : list of dictionaries
        A data field. Each element of the list corresponds to one elevation

    """
    if datatype != "VIS":
        warn("Unknown DEM data type " + datatype)
        return None

    header_size = 64
    naz = 360
    nrngs = [
        492,
        420,
        492,
        324,
        366,
        324,
        292,
        324,
        280,
        242,
        222,
        200,
        174,
        150,
        124,
        100,
        82,
        68,
        60,
        54,
    ]

    try:
        with open(fname, "rb") as visfile:
            field_list = list()
            # skip header
            visfile.seek(header_size, os.SEEK_SET)

            bindata = np.fromfile(visfile, dtype="uint8", count=-1)

            sweep_start_index = 0
            for nrng in nrngs:
                sweep_end_index = sweep_start_index + naz * nrng - 1
                field_data = np.reshape(
                    bindata[sweep_start_index : sweep_end_index + 1], (naz, nrng)
                ).astype(float)
                sweep_start_index = sweep_end_index + 1

                field_name = get_fieldname_pyart(datatype)
                field = get_metadata(field_name)
                field["data"] = field_data
                field_list.append(field)

            visfile.close()

            return field_list

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_mf_vis(fname, datatype):
    """
    Reads Météo-France visibility data binary file.

    Parameters
    ----------
    fname : str
        name of the file to read

    datatype : str
        name of the data type

    Returns
    -------
    field : dict
        A data field

    """
    datatype = datatype.strip()
    if datatype != "VIS":
        warn("Unknown DEM data type " + datatype)
        return None

    try:
        with open(fname, "rb") as visfile:
            if os.path.getsize(fname) == 262144:
                bindata = np.fromfile(visfile, dtype="uint8", count=-1)
                bindata = np.reshape(bindata, (512, 512)).astype(float)
                field_data = np.ma.masked_values(bindata, 255)
                if "coefmasqZ" not in fname:
                    field_data = 100.0 * np.ma.power(field_data / 100.0, 1.6)
                field_data = 100.0 - np.broadcast_to(field_data[::-1, :], (1, 512, 512))
            else:
                bindata = 100.0 - np.fromfile(visfile, dtype="float32", count=-1)
                field_data = np.reshape(bindata, (720, 270))

            field_name = get_fieldname_pyart(datatype)
            field = get_metadata(field_name)
            field["data"] = field_data

            visfile.close()

            return field

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_histogram(fname):
    """
    Reads a histogram contained in a csv file

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    hist , bin_edges : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            bin_edges = np.zeros(nrows + 1, dtype=float)
            hist = np.zeros(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                bin_edges[i] = float(row["bin_edge_left"])
                bin_edges[i + 1] = float(row["bin_edge_right"])
                hist[i] = float(row["value"])

            return hist, bin_edges
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None


def read_quantiles(fname):
    """
    Reads quantiles contained in a csv file

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    quantiles, values : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            quantiles = np.zeros(nrows, dtype=float)
            values = np.zeros(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                quantiles[i] = float(row["quantile"])
                values[i] = float(row["value"])

            return quantiles, values
    except (EnvironmentError, ValueError) as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None


def read_excess_gates(fname):
    """
    Reads a csv files containing the position of gates exceeding
    a given percentile of frequency of occurrence

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    rad1_ray_ind, rad1_rng_ind, rad1_ele, rad1_azi, rad1_rng,
    rad2_ray_ind, rad2_rng_ind, rad2_ele, rad2_azi, rad2_rng : tupple
        A tupple with the data read. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            ray_ind = np.empty(nrows, dtype=int)
            rng_ind = np.empty(nrows, dtype=int)
            ele = np.empty(nrows, dtype=float)
            azi = np.empty(nrows, dtype=float)
            rng = np.empty(nrows, dtype=float)
            nsamples = np.empty(nrows, dtype=int)
            occurrence = np.empty(nrows, dtype=int)
            freq_occu = np.empty(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                ray_ind[i] = int(row["ray_ind"])
                rng_ind[i] = int(row["rng_ind"])
                ele[i] = float(row["ele"])
                azi[i] = float(row["azi"])
                rng[i] = float(row["rng"])
                nsamples[i] = int(row["nsamples"])
                occurrence[i] = int(row["occurrence"])
                freq_occu[i] = float(row["freq_occu"])

            csvfile.close()

            return (ray_ind, rng_ind, ele, azi, rng, nsamples, occurrence, freq_occu)
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None, None, None, None, None, None


def read_colocated_gates(fname):
    """
    Reads a csv files containing the position of colocated gates

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    rad1_ray_ind, rad1_rng_ind, rad1_ele, rad1_azi, rad1_rng,
    rad2_ray_ind, rad2_rng_ind, rad2_ele, rad2_azi, rad2_rng : tupple
        A tupple with the data read. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            rad1_ray_ind = np.empty(nrows, dtype=int)
            rad1_rng_ind = np.empty(nrows, dtype=int)
            rad1_ele = np.empty(nrows, dtype=float)
            rad1_azi = np.empty(nrows, dtype=float)
            rad1_rng = np.empty(nrows, dtype=float)
            rad2_ray_ind = np.empty(nrows, dtype=int)
            rad2_rng_ind = np.empty(nrows, dtype=int)
            rad2_ele = np.empty(nrows, dtype=float)
            rad2_azi = np.empty(nrows, dtype=float)
            rad2_rng = np.empty(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                rad1_ray_ind[i] = int(row["rad1_ray_ind"])
                rad1_rng_ind[i] = int(row["rad1_rng_ind"])
                rad1_ele[i] = float(row["rad1_ele"])
                rad1_azi[i] = float(row["rad1_azi"])
                rad1_rng[i] = float(row["rad1_rng"])
                rad2_ray_ind[i] = int(row["rad2_ray_ind"])
                rad2_rng_ind[i] = int(row["rad2_rng_ind"])
                rad2_ele[i] = float(row["rad2_ele"])
                rad2_azi[i] = float(row["rad2_azi"])
                rad2_rng[i] = float(row["rad2_rng"])

            csvfile.close()

            return (
                rad1_ray_ind,
                rad1_rng_ind,
                rad1_ele,
                rad1_azi,
                rad1_rng,
                rad2_ray_ind,
                rad2_rng_ind,
                rad2_ele,
                rad2_azi,
                rad2_rng,
            )
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None, None, None, None, None, None, None, None


def read_colocated_data(fname):
    """
    Reads a CSV file containing colocated data using pandas.

    Parameters
    ----------
    fname : str
        Path of time series file.

    Returns
    -------
    tuple
        A tuple with the data read:
        (rad1_time, rad1_ray_ind, rad1_rng_ind, rad1_ele, rad1_azi, rad1_rng,
        rad1_val, rad2_time, rad2_ray_ind, rad2_rng_ind, rad2_ele, rad2_azi,
        rad2_rng, rad2_val). Returns None for all if unable to read the file.
    """
    try:
        # Read the CSV file, skipping lines starting with "#"
        df = pd.read_csv(fname, comment="#", na_values="--")

        # Convert rad1_time and rad2_time to datetime
        df["rad1_time"] = pd.to_datetime(df["rad1_time"], format="%Y%m%d%H%M%S")
        df["rad2_time"] = pd.to_datetime(df["rad2_time"], format="%Y%m%d%H%M%S")

        # Create masked arrays
        def to_masked_array(column, dtype=None):
            return np.ma.masked_invalid(df[column].to_numpy(dtype=dtype))

        return (
            to_masked_array("rad1_time"),
            to_masked_array("rad1_ray_ind", dtype=int),
            to_masked_array("rad1_rng_ind", dtype=int),
            to_masked_array("rad1_ele", dtype=float),
            to_masked_array("rad1_azi", dtype=float),
            to_masked_array("rad1_rng", dtype=float),
            to_masked_array("rad1_val", dtype=float),
            to_masked_array("rad2_time"),
            to_masked_array("rad2_ray_ind", dtype=int),
            to_masked_array("rad2_rng_ind", dtype=int),
            to_masked_array("rad2_ele", dtype=float),
            to_masked_array("rad2_azi", dtype=float),
            to_masked_array("rad2_rng", dtype=float),
            to_masked_array("rad2_val", dtype=float),
        )

    except Exception as e:
        warn(str(e))
        warn("Unable to read file " + fname)
        return (None,) * 14


def read_colocated_data_time_avg(fname):
    """
    Reads a CSV file containing time-averaged colocated data using pandas.

    Parameters
    ----------
    fname : str
        Path of time series file.

    Returns
    -------
    tuple
        A tuple with the data read:
        (rad1_time, rad1_ray_ind, rad1_rng_ind, rad1_ele, rad1_azi, rad1_rng,
        rad1_dBZavg, rad1_PhiDPavg, rad1_Flagavg, rad2_time, rad2_ray_ind,
        rad2_rng_ind, rad2_ele, rad2_azi, rad2_rng, rad2_dBZavg, rad2_PhiDPavg,
        rad2_Flagavg). Returns None for all if unable to read the file.
    """
    try:
        # Read the CSV file, skipping lines starting with "#"
        df = pd.read_csv(fname, comment="#", na_values="--")

        # Convert rad1_time and rad2_time to datetime
        df["rad1_time"] = pd.to_datetime(df["rad1_time"], format="%Y%m%d%H%M%S")
        df["rad2_time"] = pd.to_datetime(df["rad2_time"], format="%Y%m%d%H%M%S")

        # Create masked arrays
        def to_masked_array(column, dtype=None):
            return np.ma.masked_invalid(df[column].to_numpy(dtype=dtype))

        return (
            to_masked_array("rad1_time"),
            to_masked_array("rad1_ray_ind", dtype=int),
            to_masked_array("rad1_rng_ind", dtype=int),
            to_masked_array("rad1_ele", dtype=float),
            to_masked_array("rad1_azi", dtype=float),
            to_masked_array("rad1_rng", dtype=float),
            to_masked_array("rad1_dBZavg", dtype=float),
            to_masked_array("rad1_PhiDPavg", dtype=float),
            to_masked_array("rad1_Flagavg", dtype=float),
            to_masked_array("rad2_time"),
            to_masked_array("rad2_ray_ind", dtype=int),
            to_masked_array("rad2_rng_ind", dtype=int),
            to_masked_array("rad2_ele", dtype=float),
            to_masked_array("rad2_azi", dtype=float),
            to_masked_array("rad2_rng", dtype=float),
            to_masked_array("rad2_dBZavg", dtype=float),
            to_masked_array("rad2_PhiDPavg", dtype=float),
            to_masked_array("rad2_Flagavg", dtype=float),
        )

    except Exception as e:
        warn(str(e))
        warn("Unable to read file " + fname)
        return (None,) * 18


def read_timeseries(fname):
    """
    Reads a time series contained in a CSV file using pandas.

    Parameters
    ----------
    fname : str
        Path of time series file

    Returns
    -------
    date, value : tuple
        A list of datetime objects and a numpy masked array of values.
    """
    try:
        # Read CSV while skipping commented lines
        df = pd.read_csv(
            fname,
            comment='#',
            parse_dates=['date'],
            usecols=['date', 'value']
        )

        # Drop duplicate timestamps
        df = df.drop_duplicates(subset='date')

        masked_values = pd.to_numeric(df['value'], errors='coerce')
        # Mask fill values
        fill_val = get_fillvalue()
        masked_values = np.ma.masked_values(masked_values.to_numpy(), fill_val)

        return df['date'].tolist(), masked_values

    except Exception as e:
        warn(str(e))
        warn('Unable to read file ' + fname)
        return None, None
    
def read_ts_cum(fname):
    """
    Reads a time series of precipitation accumulation contained in a csv file

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    date, np_radar, radar_value, np_sensor, sensor_value : tupple
        The data read

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)
            np_radar = np.zeros(nrows, dtype=int)
            radar_value = np.ma.empty(nrows, dtype=float)
            np_sensor = np.zeros(nrows, dtype=int)
            sensor_value = np.ma.empty(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            date = list()
            for i, row in enumerate(reader):
                date.append(
                    datetime.datetime.strptime(row["date"], "%Y-%m-%d %H:%M:%S")
                )
                np_radar[i] = int(row["np_radar"])
                radar_value[i] = float(row["radar_value"])
                np_sensor[i] = int(row["np_sensor"])
                sensor_value[i] = float(row["sensor_value"])

            radar_value = np.ma.masked_values(radar_value, get_fillvalue())
            sensor_value = np.ma.masked_values(sensor_value, get_fillvalue())

            csvfile.close()

            return date, np_radar, radar_value, np_sensor, sensor_value
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None, None, None


def read_ml_ts(fname):
    """
    Reads a melting layer time series contained in a csv file

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    dt_ml, ml_top_avg, ml_top_std, thick_avg, thick_std, nrays_valid,
    nrays_total : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)

            dt_ml = np.empty(nrows, dtype=datetime.datetime)
            ml_top_avg = np.ma.empty(nrows, dtype=float)
            ml_top_std = np.ma.empty(nrows, dtype=float)
            thick_avg = np.ma.empty(nrows, dtype=float)
            thick_std = np.ma.empty(nrows, dtype=float)
            nrays_valid = np.zeros(nrows, dtype=int)
            nrays_total = np.zeros(nrows, dtype=int)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))

            for i, row in enumerate(reader):
                dt_ml[i] = datetime.datetime.strptime(
                    row["date-time [UTC]"], "%Y-%m-%d %H:%M:%S"
                )
                ml_top_avg[i] = float(row["mean ml top height [m MSL]"])
                ml_top_std[i] = float(row["std ml top height [m MSL]"])
                thick_avg[i] = float(row["mean ml thickness [m]"])
                thick_std[i] = float(row["std ml thickness [m]"])
                nrays_valid[i] = int(row["N valid rays"])
                nrays_total[i] = int(row["rays total"])

            ml_top_avg = np.ma.masked_values(ml_top_avg, get_fillvalue())
            ml_top_std = np.ma.masked_values(ml_top_std, get_fillvalue())
            thick_avg = np.ma.masked_values(thick_avg, get_fillvalue())
            thick_std = np.ma.masked_values(thick_std, get_fillvalue())

            return (
                dt_ml,
                ml_top_avg,
                ml_top_std,
                thick_avg,
                thick_std,
                nrays_valid,
                nrays_total,
            )
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None, None, None, None, None


def read_monitoring_ts(quantiles, fname, sort_by_date=False):
    """
    Reads a monitoring time series contained in a CSV file using pandas.

    Parameters
    ----------
    quantiles : 1D ndarray
        The computed quantiles.
    fname : str
        Path of time series file.
    sort_by_date : bool
        If True, the read data is sorted by date prior to exit.

    Returns
    -------
    date, np_t, quantile_data, geometric_mean_dB, linear_mean_dB : tuple
        - date: Array of timestamps.
        - np_t: Array of NP values.
        - quantile_data: Dictionary with quantile values, where keys are quantile names.
        - geometric_mean_dB: Mean reflectivity in dB.
        - linear_mean_dB: Mean reflectivity computed in linear units and converted back to dB.
    """
    try:
        with open(fname, "r+") as csvfile:
            lock_file(csvfile)

            # Read the CSV file, ignoring comment lines
            df = pd.read_csv(
                csvfile,
                comment="#",
                parse_dates=["date"],
                date_parser=lambda x: datetime.datetime.strptime(x, "%Y%m%d%H%M%S"),
            )

            if sort_by_date:
                df.sort_values(by="date", inplace=True)

            df = df.drop_duplicates(subset="date", keep="last")
            unlock_file(csvfile)

        # Extract date and NP values
        date = df["date"].dt.to_pydatetime()
        np_t = df["NP"].to_numpy(dtype=int)

        # Extract all quantile values dynamically
        quantile_data = {}
        for q in quantiles:
            column_name = f"quantile_{q}"
            if column_name in df.columns:
                quantile_data[q] = np.ma.masked_values(
                    df[column_name].to_numpy(dtype=float), get_fillvalue()
                )

        # Read mean values
        geometric_mean_dB = (
            np.ma.masked_values(df["geometric_mean_dB"].to_numpy(dtype=float), get_fillvalue())
            if "geometric_mean_dB" in df.columns else np.ma.masked
        )
        linear_mean_dB = (
            np.ma.masked_values(df["linear_mean_dB"].to_numpy(dtype=float), get_fillvalue())
            if "linear_mean_dB" in df.columns else np.ma.masked
        )

        return date, np_t, quantile_data, geometric_mean_dB, linear_mean_dB

    except Exception as e:
        warn(str(e))
        warn(f"Unable to read file {fname}")
        return None, None, None, None, None


def read_monitoring_ts_old(fname):
    """
    Reads an old format of the monitoring time series contained in a text file

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    date , np_t, central_quantile, low_quantile, high_quantile : tupple
        The read data in the current format. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=["Date [YYJJJ]", "Npoints", "Value"],
            )
            nrows = sum(1 for row in reader)
            np_t = np.zeros(nrows, dtype=int)
            central_quantile = np.ma.empty(nrows, dtype=float)
            low_quantile = np.ma.masked_all(nrows, dtype=float)
            high_quantile = np.ma.masked_all(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=["Date [YYJJJ]", "Npoints", "Value"],
            )

            date = list()
            for i, row in enumerate(reader):
                date.append(datetime.datetime.strptime(row["Date [YYJJJ]"], "%y%j"))
                np_t[i] = int(row["Npoints"])
                central_quantile[i] = float(row["Value"])

            central_quantile = np.ma.masked_invalid(central_quantile)

            csvfile.close()

            return date, np_t, central_quantile, low_quantile, high_quantile
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None, None, None, None, None


def read_intercomp_scores_ts(fname, sort_by_date=False):
    """
    Reads a radar intercomparison scores csv file

    Parameters
    ----------
    fname : str
        path of time series file
    sort_by_date : bool
        if True, the read data is sorted by date prior to exit

    Returns
    -------
    date_vec, np_vec, meanbias_vec, medianbias_vec, quant25bias_vec,
    quant75bias_vec, modebias_vec, corr_vec, slope_vec, intercep_vec,
    intercep_slope1_vec : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            if FCNTL_AVAIL:
                while True:
                    try:
                        fcntl.flock(csvfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
                        break
                    except OSError as e:
                        if e.errno == errno.EAGAIN:
                            time.sleep(0.1)
                        elif e.errno == errno.EBADF:
                            warn(
                                "WARNING: No file locking is possible (NFS mount?), "
                                + "expect strange issues with multiprocessing..."
                            )
                            break
                        else:
                            raise
            else:
                while True:
                    try:
                        # Attempt to acquire an exclusive lock on the file
                        msvcrt.locking(
                            os.open(csvfile, os.O_RDWR | os.O_CREAT), msvcrt.LK_NBLCK, 0
                        )
                        break
                    except OSError as e:
                        if e.errno == 13:  # Permission denied
                            time.sleep(0.1)
                        elif e.errno == 13:  # No such file or directory
                            warn(
                                "WARNING: No file locking is possible (NFS mount?), "
                                + "expect strange issues with multiprocessing..."
                            )
                            break
                        else:
                            raise

            # first count the lines
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            nrows = sum(1 for row in reader)

            np_vec = np.zeros(nrows, dtype=int)
            meanbias_vec = np.ma.empty(nrows, dtype=float)
            medianbias_vec = np.ma.empty(nrows, dtype=float)
            quant25bias_vec = np.ma.empty(nrows, dtype=float)
            quant75bias_vec = np.ma.empty(nrows, dtype=float)
            modebias_vec = np.ma.empty(nrows, dtype=float)
            corr_vec = np.ma.empty(nrows, dtype=float)
            slope_vec = np.ma.empty(nrows, dtype=float)
            intercep_vec = np.ma.empty(nrows, dtype=float)
            intercep_slope1_vec = np.ma.empty(nrows, dtype=float)
            date_vec = np.empty(nrows, dtype=datetime.datetime)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(row for row in csvfile if not row.startswith("#"))
            for i, row in enumerate(reader):
                date_vec[i] = datetime.datetime.strptime(row["date"], "%Y%m%d%H%M%S")
                np_vec[i] = int(row["NP"])
                meanbias_vec[i] = float(row["mean_bias"])
                medianbias_vec[i] = float(row["median_bias"])
                quant25bias_vec[i] = float(row["quant25_bias"])
                quant75bias_vec[i] = float(row["quant75_bias"])
                modebias_vec[i] = float(row["mode_bias"])
                corr_vec[i] = float(row["corr"])
                slope_vec[i] = float(row["slope_of_linear_regression"])
                intercep_vec[i] = float(row["intercep_of_linear_regression"])
                intercep_slope1_vec[i] = float(
                    row["intercep_of_linear_regression_of_slope_1"]
                )

            fcntl.flock(csvfile, fcntl.LOCK_UN)
            csvfile.close()

            meanbias_vec = np.ma.masked_values(meanbias_vec, get_fillvalue())
            medianbias_vec = np.ma.masked_values(medianbias_vec, get_fillvalue())
            quant25bias_vec = np.ma.masked_values(quant25bias_vec, get_fillvalue())
            quant75bias_vec = np.ma.masked_values(quant75bias_vec, get_fillvalue())
            modebias_vec = np.ma.masked_values(modebias_vec, get_fillvalue())
            corr_vec = np.ma.masked_values(corr_vec, get_fillvalue())
            slope_vec = np.ma.masked_values(slope_vec, get_fillvalue())
            intercep_vec = np.ma.masked_values(intercep_vec, get_fillvalue())
            intercep_slope1_vec = np.ma.masked_values(
                intercep_slope1_vec, get_fillvalue()
            )

            if sort_by_date:
                ind = np.argsort(date_vec)
                date_vec = date_vec[ind]
                np_vec = np_vec[ind]
                meanbias_vec = meanbias_vec[ind]
                medianbias_vec = medianbias_vec[ind]
                quant25bias_vec = quant25bias_vec[ind]
                quant75bias_vec = quant75bias_vec[ind]
                modebias_vec = modebias_vec[ind]
                corr_vec = corr_vec[ind]
                slope_vec = slope_vec[ind]
                intercep_vec = intercep_vec[ind]
                intercep_slope1_vec = intercep_slope1_vec[ind]

            return (
                date_vec,
                np_vec,
                meanbias_vec,
                medianbias_vec,
                quant25bias_vec,
                quant75bias_vec,
                modebias_vec,
                corr_vec,
                slope_vec,
                intercep_vec,
                intercep_slope1_vec,
            )
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return (None, None, None, None, None, None, None, None, None, None, None)


def read_intercomp_scores_ts_old(fname):
    """
    Reads a radar intercomparison scores csv file in old format

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    date_vec, np_vec, meanbias_vec, medianbias_vec, quant25bias_vec,
    quant75bias_vec, modebias_vec, corr_vec, slope_vec, intercep_vec,
    intercep_slope1_vec : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=[
                    "date",
                    "NP",
                    "mode_bias",
                    "median_bias",
                    "mean_bias",
                    "corr",
                    "slope_of_linear_regression",
                    "intercep_of_linear_regression",
                ],
                dialect="excel-tab",
            )
            nrows = sum(1 for row in reader)

            np_vec = np.zeros(nrows, dtype=int)
            meanbias_vec = np.ma.empty(nrows, dtype=float)
            medianbias_vec = np.ma.empty(nrows, dtype=float)
            quant25bias_vec = np.ma.masked_all(nrows, dtype=float)
            quant75bias_vec = np.ma.masked_all(nrows, dtype=float)
            modebias_vec = np.ma.empty(nrows, dtype=float)
            corr_vec = np.ma.empty(nrows, dtype=float)
            slope_vec = np.ma.empty(nrows, dtype=float)
            intercep_vec = np.ma.empty(nrows, dtype=float)
            intercep_slope1_vec = np.ma.masked_all(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=[
                    "date",
                    "NP",
                    "mode_bias",
                    "median_bias",
                    "mean_bias",
                    "corr",
                    "slope_of_linear_regression",
                    "intercep_of_linear_regression",
                ],
                dialect="excel-tab",
            )
            date_vec = list()
            for i, row in enumerate(reader):
                date_vec.append(datetime.datetime.strptime(row["date"], "%y%j"))
                np_vec[i] = int(row["NP"])
                meanbias_vec[i] = float(row["mean_bias"])
                medianbias_vec[i] = float(row["median_bias"])
                modebias_vec[i] = float(row["mode_bias"])
                corr_vec[i] = float(row["corr"])
                slope_vec[i] = float(row["slope_of_linear_regression"])
                intercep_vec[i] = float(row["intercep_of_linear_regression"])

            meanbias_vec = np.ma.masked_invalid(meanbias_vec)
            medianbias_vec = np.ma.masked_invalid(medianbias_vec)
            modebias_vec = np.ma.masked_invalid(modebias_vec)
            corr_vec = np.ma.masked_invalid(corr_vec)
            slope_vec = np.ma.masked_invalid(slope_vec)
            intercep_vec = np.ma.masked_invalid(intercep_vec)

            csvfile.close()

            return (
                date_vec,
                np_vec,
                meanbias_vec,
                medianbias_vec,
                quant25bias_vec,
                quant75bias_vec,
                modebias_vec,
                corr_vec,
                slope_vec,
                intercep_vec,
                intercep_slope1_vec,
            )
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return (None, None, None, None, None, None, None, None, None, None, None)


def read_intercomp_scores_ts_old_v0(fname, corr_min=0.6, np_min=9):
    """
    Reads a radar intercomparison scores csv file in the oldest format

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    date_vec, np_vec, meanbias_vec, medianbias_vec, quant25bias_vec,
    quant75bias_vec, modebias_vec, corr_vec, slope_vec, intercep_vec,
    intercep_slope1_vec : tupple
        The read data. None otherwise

    """
    try:
        with open(fname, "r", newline="") as csvfile:
            # first count the lines
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=["date", "ele", "NP", "mean_bias", "corr"],
                dialect="excel-tab",
            )
            nrows = sum(1 for row in reader)

            np_aux_vec = np.zeros(nrows, dtype=int)
            meanbias_aux_vec = np.ma.empty(nrows, dtype=float)
            corr_aux_vec = np.ma.empty(nrows, dtype=float)

            # now read the data
            csvfile.seek(0)
            reader = csv.DictReader(
                (row for row in csvfile if not row.startswith("#")),
                fieldnames=["date", "ele", "NP", "mean_bias", "corr"],
                dialect="excel-tab",
            )
            date_aux_vec = list()
            for i, row in enumerate(reader):
                date_aux_vec.append(datetime.datetime.strptime(row["date"], "%y%j"))
                np_aux_vec[i] = int(row["NP"])
                meanbias_aux_vec[i] = float(row["mean_bias"])
                corr_aux_vec[i] = float(row["corr"])

            meanbias_aux_vec = np.ma.masked_invalid(meanbias_aux_vec)
            corr_aux_vec = np.ma.masked_invalid(corr_aux_vec)

            csvfile.close()

            date_aux_vec = np.asarray(date_aux_vec)
            date_vec, unique_ind = np.unique(date_aux_vec, return_index=True)
            nelements = len(date_vec)

            np_vec = np.zeros(nelements, dtype=int)
            meanbias_vec = np.ma.empty(nelements, dtype=float)
            corr_vec = np.ma.empty(nelements, dtype=float)
            for i in range(nelements - 1):
                ind_aux = np.arange(unique_ind[i], unique_ind[i + 1])
                np_aux = np_aux_vec[ind_aux]
                meanbias_aux = meanbias_aux_vec[ind_aux]
                corr_aux = corr_aux_vec[ind_aux]

                ind_valid = np.where(
                    np.logical_and(corr_aux > corr_min, np_aux > np_min)
                )
                np_aux = np_aux[ind_valid]
                meanbias_aux = meanbias_aux[ind_valid]
                corr_aux = corr_aux[ind_valid]

                np_vec[i] = np.sum(np_aux, dtype=int)
                if np_vec[i] == 0:
                    continue
                meanbias_vec[i] = np.sum(np_aux * meanbias_aux) / np_vec[i]
                corr_vec[i] = np.sum(np_aux * corr_aux) / np_vec[i]

            # last date
            ind_aux = np.arange(unique_ind[-1], len(date_aux_vec))
            np_aux = np_aux_vec[ind_aux]
            meanbias_aux = meanbias_aux_vec[ind_aux]
            corr_aux = corr_aux_vec[ind_aux]

            ind_valid = np.where(np.logical_and(corr_aux > corr_min, np_aux > np_min))
            np_aux = np_aux[ind_valid]
            meanbias_aux = meanbias_aux[ind_valid]
            corr_aux = corr_aux[ind_valid]
            np_vec[-1] = np.sum(np_aux, dtype=int)
            if np_vec[-1] > 0:
                meanbias_vec[-1] = np.sum(np_aux * meanbias_aux) / np_vec[-1]
                corr_vec[-1] = np.sum(np_aux * corr_aux) / np_vec[-1]

            medianbias_vec = np.ma.masked_all(nelements, dtype=float)
            quant25bias_vec = np.ma.masked_all(nelements, dtype=float)
            quant75bias_vec = np.ma.masked_all(nelements, dtype=float)
            modebias_vec = np.ma.masked_all(nelements, dtype=float)
            slope_vec = np.ma.masked_all(nelements, dtype=float)
            intercep_vec = np.ma.masked_all(nelements, dtype=float)
            intercep_slope1_vec = np.ma.masked_all(nelements, dtype=float)

            return (
                date_vec,
                np_vec,
                meanbias_vec,
                medianbias_vec,
                quant25bias_vec,
                quant75bias_vec,
                modebias_vec,
                corr_vec,
                slope_vec,
                intercep_vec,
                intercep_slope1_vec,
            )
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return (None, None, None, None, None, None, None, None, None, None, None)


def read_selfconsistency(fname):
    """
    Reads a self-consistency table with Zdr, Kdp/Zh columns

    Parameters
    ----------
    fname : str
        path of time series file

    Returns
    -------
    zdr, kdpzh : arrays
        The read values

    """
    try:
        with open(fname, "r", newline="", encoding="utf-8", errors="ignore") as csvfile:
            # first count the lines
            reader = csv.DictReader(
                csvfile, fieldnames=["zdr", "kdpzh"], dialect="excel-tab"
            )
            nrows = sum(1 for row in reader)

            zdr_kdpzh_table = np.empty((2, nrows), dtype="float32")

            # now read the data
            csvfile.seek(0)

            reader = csv.DictReader(
                csvfile, fieldnames=["zdr", "kdpzh"], dialect="excel-tab"
            )
            for i, row in enumerate(reader):
                zdr_kdpzh_table[0, i] = float(row["zdr"])
                zdr_kdpzh_table[1, i] = float(row["kdpzh"])

            csvfile.close()

            return zdr_kdpzh_table
    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_antenna_pattern(fname, linear=False, twoway=False):
    """
    Read antenna pattern from file

    Parameters
    ----------
    fname : str
        path of the antenna pattern file
    linear : boolean
        if true the antenna pattern is given in linear units
    twoway : boolean
        if true the attenuation is two-way

    Returns
    -------
    pattern : dict
        dictionary with the fields angle and attenuation

    """

    try:
        pfile = open(fname, "r")
    except EnvironmentError as ee:
        warn(str(ee))
        raise Exception("ERROR: Could not find|open antenna file '" + fname + "'")

    # Get array length
    linenum = 0
    for line in pfile:
        if line.startswith("# CRC:"):
            break
        linenum += 1

    pattern = {"angle": np.empty(linenum), "attenuation": np.empty(linenum)}

    pfile.seek(0)
    cnt = 0
    for line in pfile:
        if line.startswith("# CRC:"):
            break
        line = line.strip()
        fields = line.split(",")
        pattern["angle"][cnt] = float(fields[0])
        pattern["attenuation"][cnt] = float(fields[1])
        cnt += 1

    pfile.close()

    if twoway:
        pattern["attenuation"] = 2.0 * pattern["attenuation"]

    if linear:
        pattern["attenuation"] = 10.0 ** (pattern["attenuation"] / 10.0)

    return pattern


def read_xls(xls_file):
    """Reads an excel file such as those used for CPC vlaidation

    Parameters
    ----------
    fname : str
        Full path of the excel file to be read

    Returns
    -------
    The excel file as a Pandas dataframe
    """

    data = pd.read_excel(xls_file, sheet_name=None)
    keys = list(data.keys())
    hourly_keys = []
    for k in keys:
        if "Data Hourly" in k:
            hourly_keys.append(k)
    out = pd.concat([data[k] for k in hourly_keys])
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
================================================
movie_maker
================================================

This program produces a movie and a gif out of all files present in a folder

"""

# Author: fvj
# License: BSD 3 clause

import glob
import os
import datetime
import atexit

from moviepy.editor import ImageSequenceClip

print(__doc__)


def main():
    """
    main programme
    """
    print(
        "====== Movie maker started: %s"
        % datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )
    atexit.register(_print_end_msg, "====== Movie maker finished: ")

    file_type = "png"
    movie_type = "mp4"
    codec = "mpeg4"
    frames_per_second = 1
    movie_path = "/utemp/mdso/figuerasiventuraj/movies/"
    file_path_list = [
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBuZ/PPI_EL00/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBuZ/PPI_EL01/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBuZ/PPI_EL02/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBuZ/PPI_EL03/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBZ/PPI_EL00/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBZ/PPI_EL01/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBZ/PPI_EL02/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/dBZ/PPI_EL03/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/hydroMF_oper/PPI_EL00/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/hydroMF_oper/PPI_EL01/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/hydroMF_oper/PPI_EL02/",
        "/utemp/mdso/figuerasiventuraj/pyrad_products/MF_ODIM_OPOU_HAIL/2020-07-01/hydroMF_oper/PPI_EL03/",
    ]
    movie_name_list = [
        "20200701_OPOU_ppi_RAW_dBuZ_el0.6",
        "20200701_OPOU_ppi_RAW_dBuZ_el1.0",
        "20200701_OPOU_ppi_RAW_dBuZ_el1.4",
        "20200701_OPOU_ppi_RAW_dBuZ_el1.8",
        "20200701_OPOU_ppi_RAW_dBZ_el0.6",
        "20200701_OPOU_ppi_RAW_dBZ_el1.0",
        "20200701_OPOU_ppi_RAW_dBZ_el1.4",
        "20200701_OPOU_ppi_RAW_dBZ_el1.8",
        "20200701_OPOU_ppi_RAW_hydroMF_oper_el0.6",
        "20200701_OPOU_ppi_RAW_hydroMF_oper_el1.0",
        "20200701_OPOU_ppi_RAW_hydroMF_oper_el1.4",
        "20200701_OPOU_ppi_RAW_hydroMF_oper_el1.8",
    ]

    create_movie(
        file_path_list,
        movie_name_list,
        movie_path,
        file_type=file_type,
        fps=frames_per_second,
        movie_type=movie_type,
        codec=codec,
    )


def create_movie(
    file_path_list,
    movie_name_list,
    movie_path,
    file_type="png",
    fps=1,
    movie_type="mp4",
    codec="mpeg4",
):
    """
    creates the movie.

    can support any type supported by ffmpeg
    some examples:
    movie type / codec
    .avi / rawvideo, png
    .mp4 / libx264, mpeg4
    avi/rawvideo supported by libreoffice
    mp4 supported by windows media player

    Parameters
    ----------
    file_path_list : list of str
        List of folders where to find the images for the movies
    movie_name_list : list of str
        List of movies to create_movie
    movie_path : str
        path where to store the movies
    file_type : str
        the individual images file type
    fps : int
        the frames per second
    movie_type : str
        the type of movie file
    codec : str
        the codec used for the movie

    Returns
    -------
    Nothing

    """
    for movie_name, file_path in zip(movie_name_list, file_path_list):
        file_list = sorted(glob.glob(file_path + "*." + file_type))
        print(file_list)

        # Generate clip
        clip = ImageSequenceClip(file_list, fps=fps)
        # Write out clip
        if not os.path.isdir(movie_path):
            os.makedirs(movie_path)
        clip.write_videofile(movie_path + movie_name + "." + movie_type, codec=codec)
        clip.write_gif(movie_path + movie_name + ".gif")

        print("Created movie " + movie_path + movie_name)


def _print_end_msg(text):
    """
    prints end message

    Parameters
    ----------
    text : str
        the text to be printed

    Returns
    -------
    Nothing

    """
    print(
        text
        + datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    )


# ---------------------------------------------------------
# Start main:
# ---------------------------------------------------------
if __name__ == "__main__":
    main()



"""
pyrad.proc.process_spectra
==========================

Functions to processes spectral data.

.. autosummary::
    :toctree: generated/

    process_raw_spectra
    process_ifft
    process_spectra_point
    process_filter_0Doppler
    process_filter_srhohv
    process_filter_spectra_noise
    process_dealias_spectra
    process_spectra_ang_avg
    process_spectral_power
    process_spectral_noise
    process_spectral_phase
    process_spectral_reflectivity
    process_spectral_differential_reflectivity
    process_spectral_differential_phase
    process_spectral_rhohv
    process_pol_variables
    process_noise_power
    process_reflectivity
    process_differential_reflectivity
    process_differential_phase
    process_rhohv
    process_Doppler_velocity
    process_Doppler_width

"""

from copy import deepcopy
from warnings import warn
import numpy as np
from netCDF4 import num2date

import pyart

from ..io.io_aux import get_datatype_fields, get_fieldname_pyart


def process_raw_spectra(procstatus, dscfg, radar_list=None):
    """
    Dummy function that returns the initial input data set

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break
    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    new_dataset = {"radar_out": deepcopy(radar_list[ind_rad])}

    return new_dataset, ind_rad


def process_ifft(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectrum width from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "ShhADUu", or,
            "SvvADU" or "SvvADUu", or,
            "sNADUh" or "sNADUv"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "IQhhADU" if "ShhADU" or "ShhADUu" were provided,
        "IQhvvDU" if "SvvADU" or "SvvADUu" were provided,
        "IQNADUh" if "sNADUh" was provided,
        "IQNADUv" if "sNADUv" was provided.
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    wind_params = dscfg.get("window", ["None"])
    if len(wind_params) == 1:
        window = wind_params[0]
        if window == "None":
            window = None
        else:
            try:
                window = float(window)
            except ValueError:
                pass
    else:
        window = wind_params
        for i in range(1, len(window)):
            window[i] = float(window[i])
        window = tuple(window)

    fields_in_list = []
    fields_out_list = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name = get_fieldname_pyart(datatype)
        if field_name not in radar.fields:
            warn(field_name + " not in radar")
            continue
        if field_name in (
            "unfiltered_complex_spectra_hh_ADU",
            "complex_spectra_hh_ADU",
        ):
            fields_out_list.append("IQ_hh_ADU")
        elif field_name in (
            "unfiltered_complex_spectra_vv_ADU",
            "complex_spectra_vv_ADU",
        ):
            fields_out_list.append("IQ_vv_ADU")
        elif field_name == "spectral_noise_power_hh_ADU":
            fields_out_list.append("IQ_noise_power_hh_ADU")
        elif field_name == "spectral_noise_power_vv_ADU":
            fields_out_list.append("IQ_noise_power_vv_ADU")
        else:
            warn(field_name + " can not be inverse Fourier transformed")
        fields_in_list.append(field_name)

    radar_out = pyart.retrieve.compute_iq(
        radar, fields_in_list, fields_out_list, window=window
    )

    # prepare for exit
    new_dataset = {"radar_out": radar_out}

    return new_dataset, ind_rad


def process_spectra_point(procstatus, dscfg, radar_list=None):
    """
    Obtains the spectra or IQ data at a point location.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement
        single_point : boolean. Dataset keyword
            if True only one gate per radar volume is going to be kept.
            Otherwise all gates within the azimuth and elevation tolerance
            are going to be kept. This is useful to extract all data from
            fixed pointing scans. Default True
        latlon : boolean. Dataset keyword
            if True position is obtained from latitude, longitude information,
            otherwise position is obtained from antenna coordinates
            (range, azimuth, elevation). Default False
        truealt : boolean. Dataset keyword
            if True the user input altitude is used to determine the point of
            interest.
            if False use the altitude at a given radar elevation ele over the
            point of interest. Default True
        lon : float. Dataset keyword
            the longitude [deg]. Use when latlon is True.
        lat : float. Dataset keyword
            the latitude [deg]. Use when latlon is True.
        alt : float. Dataset keyword
            altitude [m MSL]. Use when latlon is True. Default 0.
        ele : float. Dataset keyword
            radar elevation [deg]. Use when latlon is False or when latlon is
            True and truealt is False
        azi : float. Dataset keyword
            radar azimuth [deg]. Use when latlon is False
        rng : float. Dataset keyword
            range from radar [m]. Use when latlon is False
        AziTol : float. Dataset keyword
            azimuthal tolerance to determine which radar azimuth to use [deg].
            Default 0.5
        EleTol : float. Dataset keyword
            elevation tolerance to determine which radar elevation to use
            [deg]. Default 0.5
        RngTol : float. Dataset keyword
            range tolerance to determine which radar bin to use [m]. Default
            50.

    radar_list : list of Radar objects
          Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the data and metadata at the point of interest
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        return None, None

    field_names = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_names.append(get_fieldname_pyart(datatype))

    ind_rad = int(radarnr[5:8]) - 1

    if procstatus == 2:
        if dscfg["initialized"] == 0:
            return None, None

        # prepare for exit
        new_dataset = {
            "radar_out": dscfg["global_data"]["psr_poi"],
            "point_coordinates_WGS84_lon_lat_alt": (
                dscfg["global_data"]["point_coordinates_WGS84_lon_lat_alt"]
            ),
            "antenna_coordinates_az_el_r": (
                dscfg["global_data"]["antenna_coordinates_az_el_r"]
            ),
            "final": True,
        }

        return new_dataset, ind_rad

    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid psr")
        return None, None
    psr = radar_list[ind_rad]

    projparams = dict()
    projparams.update({"proj": "pyart_aeqd"})
    projparams.update({"lon_0": psr.longitude["data"]})
    projparams.update({"lat_0": psr.latitude["data"]})

    truealt = dscfg.get("truealt", True)
    latlon = dscfg.get("latlon", False)

    single_point = dscfg.get("single_point", True)
    if latlon:
        lon = dscfg["lon"]
        lat = dscfg["lat"]
        alt = dscfg.get("alt", 0.0)
        dscfg.get("latlonTol", 1.0)
        dscfg.get("altTol", 100.0)

        x, y = pyart.core.geographic_to_cartesian(lon, lat, projparams)

        if not truealt:
            ke = 4.0 / 3.0  # constant for effective radius
            a = 6378100.0  # earth radius
            re = a * ke  # effective radius

            elrad = dscfg["ele"] * np.pi / 180.0
            r_ground = np.sqrt(x**2.0 + y**2.0)
            r = r_ground / np.cos(elrad)
            alt_psr = (
                psr.altitude["data"]
                + np.sqrt(r**2.0 + re**2.0 + 2.0 * r * re * np.sin(elrad))
                - re
            )
            alt_psr = alt_psr[0]
        else:
            alt_psr = alt

        r, az, el = pyart.core.cartesian_to_antenna(
            x, y, alt_psr - psr.altitude["data"]
        )
        r = r[0]
        az = az[0]
        el = el[0]
    else:
        r = dscfg["rng"]
        az = dscfg["azi"]
        el = dscfg["ele"]
        azi_tol = dscfg.get("AziTol", 0.5)
        ele_tol = dscfg.get("EleTol", 0.5)
        rng_tol = dscfg.get("RngTol", 50.0)

        x, y, alt = pyart.core.antenna_to_cartesian(r / 1000.0, az, el)
        lon, lat = pyart.core.cartesian_to_geographic(x, y, projparams)
        lon = lon[0]
        lat = lat[0]

    d_az = np.abs(psr.azimuth["data"] - az)
    if np.min(d_az) > azi_tol:
        warn(
            " No psr bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to psr azimuth "
            + str(d_az)
            + " larger than tolerance"
        )
        return None, None

    d_el = np.abs(psr.elevation["data"] - el)
    if np.min(d_el) > ele_tol:
        warn(
            " No psr bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to psr elevation "
            + str(d_el)
            + " larger than tolerance"
        )
        return None, None

    d_r = np.abs(psr.range["data"] - r)
    if np.min(d_r) > rng_tol:
        warn(
            " No psr bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to psr range bin "
            + str(d_r)
            + " larger than tolerance"
        )
        return None, None

    if single_point:
        ind_ray = np.argmin(
            np.abs(psr.azimuth["data"] - az) + np.abs(psr.elevation["data"] - el)
        )
    else:
        ind_ray = np.where(np.logical_and(d_az <= azi_tol, d_el <= ele_tol))[0]
    ind_rng = np.argmin(np.abs(psr.range["data"] - r))
    nrays = ind_ray.size

    time_poi = num2date(
        psr.time["data"][ind_ray], psr.time["units"], psr.time["calendar"]
    )
    if nrays > 1:
        time_ref = time_poi[0]
    else:
        time_ref = time_poi
        time_poi = np.array([time_poi])

    # initialize dataset
    if not dscfg["initialized"]:
        psr_poi = deepcopy(psr)

        # prepare space for field
        psr_poi.fields = dict()
        for field_name in field_names:
            if field_name not in psr.fields:
                warn("Field " + field_name + " not in psr object")
                return None, None
            psr_poi.add_field(field_name, deepcopy(psr.fields[field_name]))
            psr_poi.fields[field_name]["data"] = np.array([])

        # fixed psr objects parameters
        psr_poi.range["data"] = np.array([r])
        psr_poi.ngates = 1

        psr_poi.time["units"] = pyart.io.make_time_unit_str(time_ref)
        psr_poi.time["data"] = np.array([])
        psr_poi.scan_type = "poi_time_series"
        psr_poi.sweep_number["data"] = np.array([], dtype=np.int32)
        psr_poi.nsweeps = 1
        psr_poi.sweep_mode["data"] = np.array(["poi_time_series"])
        psr_poi.rays_are_indexed = None
        psr_poi.ray_angle_res = None
        psr_poi.fixed_angle["data"] = np.array([az])

        # ray dependent psr objects parameters
        psr_poi.sweep_end_ray_index["data"] = np.array([-1], dtype="int32")
        psr_poi.rays_per_sweep["data"] = np.array([0], dtype="int32")
        psr_poi.azimuth["data"] = np.array([], dtype="float64")
        psr_poi.elevation["data"] = np.array([], dtype="float64")
        psr_poi.nrays = 0

        psr_poi.npulses["data"] = np.array([], dtype=np.int32)
        if psr_poi.Doppler_velocity is not None:
            psr_poi.Doppler_velocity["data"] = np.array([])
        if psr_poi.Doppler_frequency is not None:
            psr_poi.Doppler_frequency["data"] = np.array([])

        dscfg["global_data"] = {
            "psr_poi": psr_poi,
            "point_coordinates_WGS84_lon_lat_alt": [lon, lat, alt],
            "antenna_coordinates_az_el_r": [az, el, r],
        }

        dscfg["initialized"] = 1

    psr_poi = dscfg["global_data"]["psr_poi"]
    psr_poi.sweep_end_ray_index["data"][0] += nrays
    psr_poi.rays_per_sweep["data"][0] += nrays
    psr_poi.nrays += nrays
    psr_poi.azimuth["data"] = np.append(psr_poi.azimuth["data"], np.zeros(nrays) + az)
    psr_poi.elevation["data"] = np.append(
        psr_poi.elevation["data"], np.zeros(nrays) + el
    )
    start_time = num2date(0, psr_poi.time["units"], psr_poi.time["calendar"])
    for i in range(nrays):
        psr_poi.time["data"] = np.append(
            psr_poi.time["data"], (time_poi[i] - start_time).total_seconds()
        )

    psr_poi.gate_longitude["data"] = (
        np.ones((psr_poi.nrays, psr_poi.ngates), dtype="float64") * lon
    )
    psr_poi.gate_latitude["data"] = (
        np.ones((psr_poi.nrays, psr_poi.ngates), dtype="float64") * lat
    )
    psr_poi.gate_altitude["data"] = np.broadcast_to(
        alt, (psr_poi.nrays, psr_poi.ngates)
    )

    for field_name in field_names:
        dtype = psr_poi.fields[field_name]["data"].dtype
        if field_name not in psr.fields:
            warn("Field " + field_name + " not in psr object")
            poi_data = np.ma.masked_all((nrays, 1, psr.npulses_max), dtype=dtype)
        else:
            poi_data = psr.fields[field_name]["data"][ind_ray, ind_rng, :]
            poi_data = poi_data.reshape(nrays, 1, psr.npulses_max)

        # Put data in radar object
        if np.size(psr_poi.fields[field_name]["data"]) == 0:
            psr_poi.fields[field_name]["data"] = poi_data.reshape(
                nrays, 1, psr_poi.npulses_max
            )
        else:
            if psr_poi.npulses_max == psr.npulses_max:
                psr_poi.fields[field_name]["data"] = np.ma.append(
                    psr_poi.fields[field_name]["data"], poi_data, axis=0
                )
            elif psr.npulses_max < psr_poi.npulses_max:
                poi_data_aux = np.ma.masked_all(
                    (nrays, 1, psr_poi.npulses_max), dtype=dtype
                )
                for i in range(nrays):
                    poi_data_aux[i, 0, 0 : psr.npulses_max] = poi_data[i, 0, :]
                psr_poi.fields[field_name]["data"] = np.ma.append(
                    psr_poi.fields[field_name]["data"], poi_data_aux, axis=0
                )
            else:
                poi_data_aux = np.ma.masked_all(
                    (psr_poi.nrays, 1, psr.npulses_max), dtype=dtype
                )
                poi_data_aux[: psr_poi.nrays - nrays, :, 0 : psr_poi.npulses_max] = (
                    psr_poi.fields[field_name]["data"]
                )
                poi_data_aux[psr_poi.nrays - nrays :, :, :] = poi_data
                psr_poi.fields[field_name]["data"] = poi_data_aux

    psr_poi.npulses["data"] = np.append(
        psr_poi.npulses["data"], psr.npulses["data"][ind_ray]
    )
    if psr_poi.Doppler_velocity is not None:
        if np.size(psr_poi.Doppler_velocity["data"]) == 0:
            psr_poi.Doppler_velocity["data"] = psr.Doppler_velocity["data"][
                ind_ray, :
            ].reshape(nrays, psr_poi.npulses_max)
        else:
            Doppler_data = psr.Doppler_velocity["data"][ind_ray, :]
            Doppler_data = Doppler_data.reshape(nrays, psr.npulses_max)

            if psr_poi.npulses_max == psr.npulses_max:
                psr_poi.Doppler_velocity["data"] = np.ma.append(
                    psr_poi.Doppler_velocity["data"], Doppler_data, axis=0
                )
            elif psr.npulses_max < psr_poi.npulses_max:
                Doppler_aux = np.ma.masked_all((nrays, psr_poi.npulses_max))
                for i in range(nrays):
                    Doppler_aux[i, 0 : psr.npulses_max] = Doppler_data[i, :]
                psr_poi.Doppler_velocity["data"] = np.ma.append(
                    psr_poi.Doppler_velocity["data"], Doppler_aux, axis=0
                )
            else:
                Doppler_aux = np.ma.masked_all((psr_poi.nrays, psr.npulses_max))
                Doppler_aux[: psr_poi.nrays - nrays, 0 : psr_poi.npulses_max] = (
                    psr_poi.Doppler_velocity["data"]
                )
                Doppler_aux[psr_poi.nrays - nrays :, :] = Doppler_data
                psr_poi.Doppler_velocity["data"] = Doppler_aux

    if psr_poi.Doppler_frequency is not None:
        if np.size(psr_poi.Doppler_frequency["data"]) == 0:
            psr_poi.Doppler_frequency["data"] = psr.Doppler_frequency["data"][
                ind_ray, :
            ].reshape(nrays, psr_poi.npulses_max)
        else:
            Doppler_data = psr.Doppler_frequency["data"][ind_ray, :]
            Doppler_data = Doppler_data.reshape(nrays, psr.npulses_max)

            if psr_poi.npulses_max == psr.npulses_max:
                psr_poi.Doppler_frequency["data"] = np.ma.append(
                    psr_poi.Doppler_frequency["data"], Doppler_data, axis=0
                )
            elif psr.npulses_max < psr_poi.npulses_max:
                Doppler_aux = np.ma.masked_all((nrays, psr_poi.npulses_max))
                for i in range(nrays):
                    Doppler_aux[i, 0 : psr.npulses_max] = Doppler_data[i, :]
                psr_poi.Doppler_frequency["data"] = np.ma.append(
                    psr_poi.Doppler_frequency["data"], Doppler_aux, axis=0
                )
            else:
                Doppler_aux = np.ma.masked_all((psr_poi.nrays, psr.npulses_max))
                Doppler_aux[0 : psr_poi.nrays - nrays, 0 : psr_poi.npulses_max] = (
                    psr_poi.Doppler_frequency["data"]
                )
                Doppler_aux[psr_poi.nrays - nrays :, :] = Doppler_data
                psr_poi.Doppler_frequency["data"] = Doppler_aux

    psr_poi.npulses_max = max(psr_poi.npulses_max, psr.npulses_max)

    dscfg["global_data"]["psr_poi"] = psr_poi

    # prepare for exit
    new_dataset = {
        "radar_out": psr_poi,
        "point_coordinates_WGS84_lon_lat_alt": (
            dscfg["global_data"]["point_coordinates_WGS84_lon_lat_alt"]
        ),
        "antenna_coordinates_az_el_r": (
            dscfg["global_data"]["antenna_coordinates_az_el_r"]
        ),
        "final": False,
    }

    return new_dataset, ind_rad


def process_filter_0Doppler(procstatus, dscfg, radar_list=None):
    """
    Function to filter the 0-Doppler line bin and neighbours of the
    Doppler spectra

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, can be any of the spectral fields supported by pyrad

        filter_width : float
            The Doppler filter width. Default 0.
        filter_units : str
            Can be 'm/s' or 'Hz'. Default 'm/s'
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field, the names of the output fields
        is the same as the provided datatypes, except for unfiltered fields which are renamed in the following
        "dBuZ" => "dBZ"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    field_name_list = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name_list.append(get_fieldname_pyart(datatype))

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    filter_width = dscfg.get("filter_width", 0.0)
    filter_units = dscfg.get("filter_units", "m/s")

    if filter_units == "m/s":
        axis = psr.Doppler_velocity["data"]
    else:
        axis = psr.Doppler_frequency["data"]

    fields = dict()
    for field_name in field_name_list:
        if field_name not in psr.fields:
            warn("Unable to filter 0-Doppler. Missing field " + field_name)
            continue

        field_name_aux = field_name.replace("unfiltered_", "")
        field = pyart.config.get_metadata(field_name_aux)
        field["data"] = deepcopy(psr.fields[field_name]["data"])
        for ray in range(psr.nrays):
            ind = np.ma.where(
                np.logical_and(
                    axis[ray, :] >= -filter_width / 2.0,
                    axis[ray, :] <= filter_width / 2.0,
                )
            )
            field["data"][ray, :, ind] = np.ma.masked
        fields.update({field_name_aux: field})

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    for field_name in fields.keys():
        new_dataset["radar_out"].add_field(field_name, fields[field_name])

    return new_dataset, ind_rad


def process_filter_srhohv(procstatus, dscfg, radar_list=None):
    """
    Filter Doppler spectra as a function of spectral RhoHV

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "sRhoHV" or "sRhoHVu",
            as well as any spectral field supported by pyrad
        sRhoHV_threshold : float
            Data with sRhoHV module above this threshold will be filtered.
            Default 1.
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field, the names of the output fields
        is the same as the provided datatypes, except for unfiltered fields which are renamed in the following
        "dBuZ" => "dBZ"
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    field_name_list = []
    sRhoHV_found = False
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sRhoHV", "sRhoHVu") and not sRhoHV_found:
            sRhoHV_field = get_fieldname_pyart(datatype)
            sRhoHV_found = True
        else:
            field_name_list.append(get_fieldname_pyart(datatype))

    if not sRhoHV_found:
        warn("sRhoHV field is required for sRhoHV filtering")
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sRhoHV_field not in psr.fields:
        warn("Unable to obtain apply sRhoHV filter. Missing field " + sRhoHV_field)
        return None, None

    sRhoHV_threshold = dscfg.get("sRhoHV_threshold", 0.9)
    sRhoHV = psr.fields[sRhoHV_field]["data"]

    fields = dict()
    for field_name in field_name_list:
        if field_name not in psr.fields:
            warn("Unable to filter according to sRhoHV. Missing field " + field_name)
            continue

        field_name_aux = field_name.replace("unfiltered_", "")
        field = pyart.config.get_metadata(field_name_aux)
        field["data"] = deepcopy(psr.fields[field_name]["data"])
        field["data"][np.ma.abs(sRhoHV) <= sRhoHV_threshold] = np.ma.masked
        fields.update({field_name_aux: field})

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    for field_name in fields.keys():
        new_dataset["radar_out"].add_field(field_name, fields[field_name])

    return new_dataset, ind_rad


def process_filter_spectra_noise(procstatus, dscfg, radar_list=None):
    """
    Filter the noise of the Doppler spectra by clipping any data below
    the noise level plus a margin

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu", and,
            "sNADUh" or "sNADUv"
        clipping_level : float
            The clipping level [dB above noise level]. Default 10.
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field, the names of the output fields
        is the same as the provided datatypes
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    field_name_list = []
    signal_found = False
    noise_found = False
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu") and not signal_found:
            signal_field = get_fieldname_pyart(datatype)
            signal_found = True
        elif datatype in ("sNADUh", "sNADUv") and not noise_found:
            noise_field = get_fieldname_pyart(datatype)
            noise_found = True
        else:
            field_name_list.append(get_fieldname_pyart(datatype))

    if not signal_found or not noise_found:
        warn("Signal and noise fields are required for noise filtering")
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields or noise_field not in psr.fields:
        warn("Unable to obtain apply spectral noise filter. Missing fields")
        return None, None

    clipping_level = dscfg.get("clipping_level", 10.0)

    # get Doppler bins below clipping level
    clip_pwr = psr.fields[noise_field]["data"] * np.power(10.0, 0.1 * clipping_level)

    s_pwr = pyart.retrieve.compute_spectral_power(
        psr, units="ADU", signal_field=signal_field, noise_field=noise_field
    )

    mask = np.ma.less_equal(s_pwr["data"], clip_pwr)

    # filter data
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    for field_name in field_name_list:
        if field_name not in psr.fields:
            warn("Unable to filter field " + field_name)
            continue

        new_dataset["radar_out"].add_field(field_name, psr.fields[field_name])
        new_dataset["radar_out"].fields[field_name]["data"][mask] = np.ma.masked

    return new_dataset, ind_rad


def process_dealias_spectra(procstatus, dscfg, radar_list=None):
    """
    Performs a dealiasing of spectra data, assuming at most one fold

    The method is quite simple and works in the following way at every
    radar gate

    - aliasing check
        check if there is no noise either on the left of the right of the spectrum
    - left/right tail computation
        identify left and right tail of the aliased spectrum
    - detect direction of aliasing
        compute Doppler mean velocity with right or left shift of the spectrum,
        perform the shift which minimizes the difference in Doppler velocity to
        previous range bin

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu", and,
            "sNADUh" or "sNADUv"

    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields. The output fields are the same as the original
        ones but are dealiased.
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    field_name_list = []
    signal_found = False
    noise_found = False
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu") and not signal_found:
            signal_field = get_fieldname_pyart(datatype)
            signal_found = True
        elif datatype in ("sNADUh", "sNADUv") and not noise_found:
            noise_field = get_fieldname_pyart(datatype)
            noise_found = True
        else:
            field_name_list.append(get_fieldname_pyart(datatype))

    if not signal_found or not noise_found:
        warn("Signal and noise fields are required for noise filtering")
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields or noise_field not in psr.fields:
        warn("Unable to obtain apply spectral noise filter. Missing fields")
        return None, None

    s_pwr = pyart.retrieve.compute_spectral_power(
        psr, units="ADU", signal_field=signal_field, noise_field=noise_field
    )

    # Add spectral power field
    psr.fields["spectral_power"] = s_pwr

    # Dealias dataset
    new_dataset = pyart.retrieve.spectra.dealias_spectra(
        psr, "spectral_power", field_name_list
    )

    return new_dataset, ind_rad


def process_spectra_ang_avg(procstatus, dscfg, radar_list=None):
    """
    Function to average the spectra over the rays. This function is
    intended mainly for vertically pointing scans. The function assumes
    the volume is composed of a single sweep, it averages over the number
    of rays specified by the user and produces a single ray output.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types,
            any spectral datatype supported by pyrad
        navg : int
            Number of spectra to average. If -1 all spectra will be averaged.
            Default -1.
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the same output fields as the provided datatypes
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    field_name_list = []
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name_list.append(get_fieldname_pyart(datatype))

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    navg = dscfg.get("navg", -1)

    # keep only fields of interest
    psr_aux = deepcopy(psr)
    psr_aux.fields = dict()
    for field_name in field_name_list:
        if field_name not in psr.fields:
            warn("Field " + field_name + " missing")
            continue
        psr_aux.add_field(field_name, psr.fields[field_name])

    psr_aux = pyart.util.interpol_spectra(psr_aux)

    if navg == -1:
        navg = psr.nrays
    elif navg > psr.nrays:
        warn(
            "Number of rays " + str(psr.nrays) + " smaller than number of "
            "desired spectra to average " + str(navg)
        )
        navg = psr.nrays

    for field_name in psr_aux.fields.keys():
        data_mean = np.ma.mean(
            psr_aux.fields[field_name]["data"][0:navg, :, 0 : psr_aux.npulses_max],
            axis=0,
        )
        psr_aux.fields[field_name]["data"] = np.ma.masked_all(
            (1, psr_aux.ngates, psr_aux.npulses_max),
            dtype=psr_aux.fields[field_name]["data"].dtype,
        )
        psr_aux.fields[field_name]["data"][0, :, :] = data_mean

    psr_aux.time["data"] = np.array([psr_aux.time["data"][int(navg / 2)]])
    psr_aux.azimuth["data"] = np.array([0], dtype=np.float32)
    psr_aux.elevation["data"] = np.array([psr_aux.elevation["data"][int(navg / 2)]])
    psr_aux.nrays = 1
    psr_aux.sweep_end_ray_index["data"] = np.array([0.0], dtype=np.int32)

    psr_aux.init_rays_per_sweep()
    psr_aux.init_gate_x_y_z()
    psr_aux.init_gate_longitude_latitude()
    psr_aux.init_gate_altitude()

    if psr_aux.Doppler_velocity is not None:
        psr_aux.Doppler_velocity["data"] = np.ma.expand_dims(
            psr_aux.Doppler_velocity["data"][0, :], axis=0
        )
    if psr_aux.Doppler_frequency is not None:
        psr_aux.Doppler_frequency["data"] = np.ma.expand_dims(
            psr_aux.Doppler_frequency["data"][0, :], axis=0
        )

    # prepare for exit
    new_dataset = {"radar_out": psr_aux}

    return new_dataset, ind_rad


def process_spectral_power(procstatus, dscfg, radar_list=None):
    """
    Computes the spectral power

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu", and,
            "sNADUh", or "sNADUv"
        units : str
            The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        smooth_window : int or None
            Size of the moving Gaussian smoothing window. If none no smoothing
            will be applied
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "sPhhADU" or "sPhhADUu", or
        "sPvvADU" or "sPvvADUu", or
        "sPhhdBADU" or "sPhhdBADUu", or
        "sPvvdBADU" or "sPvvdBADUu", or
        "sPhhAdBm" or "sPhhdBmu", or
        "sPvvdBm" or "sPvvdBmu",
        depending on which input datatype and units were provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu"):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ("sNADUh", "sNADUv"):
            noise_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields:
        warn("Unable to obtain spectral signal power. Missing field " + signal_field)
        return None, None

    units = dscfg.get("units", "dBADU")
    subtract_noise = dscfg.get("subtract_noise", False)
    smooth_window = dscfg.get("smooth_window", None)

    s_pwr = pyart.retrieve.compute_spectral_power(
        psr,
        units=units,
        subtract_noise=subtract_noise,
        smooth_window=smooth_window,
        signal_field=signal_field,
        noise_field=noise_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(s_pwr["standard_name"], s_pwr)

    return new_dataset, ind_rad


def process_spectral_noise(procstatus, dscfg, radar_list=None):
    """
    Computes the spectral noise

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu"
        units : str
            The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
        navg : int
            Number of spectra averaged
        rmin : int
            Range from which the data is used to estimate the noise
        nnoise_min : int
            Minimum number of samples to consider the estimated noise power
            valid
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
            "sNADUh" or
            "sNADUv" or
            "sNdBADUh" or
            "sNdBADUv" or
            "sNdBmh" or
            "sNdBmv"
            depending on which input datatype and units were provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu"):
            signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields:
        warn("Unable to obtain spectral noise power. Missing field " + signal_field)
        return None, None

    units = dscfg.get("units", "ADU")
    navg = dscfg.get("navg", 1)
    rmin = dscfg.get("rmin", 0.0)
    nnoise_min = dscfg.get("nnoise_min", 100)

    s_pwr = pyart.retrieve.compute_spectral_noise(
        psr,
        units=units,
        navg=navg,
        rmin=rmin,
        nnoise_min=nnoise_min,
        signal_field=signal_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(s_pwr["standard_name"], s_pwr)

    return new_dataset, ind_rad


def process_spectral_phase(procstatus, dscfg, radar_list=None):
    """
    Computes the spectral phase

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "SPhasehh" if "ShhADU" was provided as input
        "SPhasehhu" if "ShhADUu" was provided as input
        "SPhasevv" if "SvvADU" was provided as input
        "SPhasevvu" if "SvvADUu" was provided as input

    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu"):
            signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields:
        warn("Unable to obtain spectral phase. Missing field " + signal_field)
        return None, None

    s_phase = pyart.retrieve.compute_spectral_phase(psr, signal_field=signal_field)

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(s_phase["standard_name"], s_phase)

    return new_dataset, ind_rad


def process_spectral_reflectivity(procstatus, dscfg, radar_list=None):
    """
    Computes spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            either a combination of signal and noise
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu", and,
            "sNADUh" or "sNADUv", or
            the power signal
            "sPhhADU" or "sPvvADU" or "sPhhADUu" or "sPvvADUu"
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        smooth_window : int or None
            Size of the moving Gaussian smoothing window. If none no smoothing
            will be applied
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "sdBZ" if "ShhADU" (or "sPhhADU") was provided as input
        "sdBuZ" if "ShhADUu" (or "sPhhADUu") was provided as input
        "sdBZv" if "SvvADU" (or "sPvvADU") was provided as input
        "sdBuZv" if "SvvADUu" (or "sPvvADUu") was provided as input
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None
    noise_field = None
    signal_field = None
    pwr_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu"):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ("sNADUh", "sNADUv"):
            noise_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPhhADU", "sPvvADU", "sPhhADUu", "sPvvADUu"):
            pwr_field = get_fieldname_pyart(datatype)

    if pwr_field is None and signal_field is None:
        warn("Either signal or power fields must be specified")
        return None, None

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    compute_power = True
    if pwr_field is not None:
        compute_power = False

    if compute_power and signal_field not in psr.fields:
        warn("Unable to obtain spectral reflectivity. Missing field " + signal_field)
        return None, None

    if not compute_power and pwr_field not in psr.fields:
        warn("Unable to obtain spectral reflectivity. Missing field " + pwr_field)
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)
    smooth_window = dscfg.get("smooth_window", None)

    sdBZ = pyart.retrieve.compute_spectral_reflectivity(
        psr,
        compute_power=compute_power,
        subtract_noise=subtract_noise,
        smooth_window=smooth_window,
        pwr_field=pwr_field,
        signal_field=signal_field,
        noise_field=noise_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(sdBZ["standard_name"], sdBZ)

    return new_dataset, ind_rad


def process_spectral_differential_reflectivity(procstatus, dscfg, radar_list=None):
    """
    Computes spectral differential reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            either a combination of signal and noise
            ("ShhADU" and "SvvADU") or ("ShhADUu" and "SvvADUu"), and,
            ("sNADUh" and "sNADUv"), or
            the power signal
            ("sPhhADU" and "sPvvADU") or ("sPhhADUu" and "sPvvADUu")
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        smooth_window : int or None
            Size of the moving Gaussian smoothing window. If none no smoothing
            will be applied
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "sZDR" if "ShhADU" and "SvvADU" were provided as input
        "sZDRu" if "ShhADUu" and "SvvADUu" were provided as input
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    pwr_h_field = None
    pwr_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "ShhADUu"):
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("SvvADU", "SvvADUu"):
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUv":
            noise_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPhhADU", "sPhhADUu"):
            pwr_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPvvADU", "sPvvADUu"):
            pwr_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    compute_power = True
    if pwr_h_field is not None and pwr_v_field is not None:
        compute_power = False

    if compute_power and (
        signal_h_field not in psr.fields or signal_v_field not in psr.fields
    ):
        warn("Unable to obtain spectral differential reflectivity. " + "Missing fields")
        return None, None
    if not compute_power and (
        pwr_h_field not in psr.fields or pwr_v_field not in psr.fields
    ):
        warn("Unable to obtain spectral differential reflectivity. " + "Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)
    smooth_window = dscfg.get("smooth_window", None)

    sZDR = pyart.retrieve.compute_spectral_differential_reflectivity(
        psr,
        compute_power=compute_power,
        subtract_noise=subtract_noise,
        smooth_window=smooth_window,
        pwr_h_field=pwr_h_field,
        pwr_v_field=pwr_v_field,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(sZDR["standard_name"], sZDR)

    return new_dataset, ind_rad


def process_spectral_differential_phase(procstatus, dscfg, radar_list=None):
    """
    Computes the spectral differential phase

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            either a combination of signal and noise
            ("ShhADU" and "SvvADU") or ("ShhADUu" and "SvvADUu"), and,
            "sRhoHV" or "sRhoHVu" (Optional)
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "sPhiDP" if "ShhADU" and "SvvADU" were provided
        "sPhiDPu" if "ShhADUu" and "SvvADUu" were provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    signal_h_field = None
    signal_v_field = None
    srhohv_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "ShhADUu"):
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("SvvADU", "SvvADUu"):
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sRhoHV", "sRhoHVu"):
            srhohv_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    use_rhohv = False
    if srhohv_field is not None:
        use_rhohv = True

    if not use_rhohv and (
        signal_h_field not in psr.fields or signal_v_field not in psr.fields
    ):
        warn("Unable to obtain spectral signal differential phase. " + "Missing fields")
        return None, None
    if use_rhohv and srhohv_field not in psr.fields:
        warn("Unable to obtain spectral signal differential phase. " + "Missing fields")
        return None, None

    sPhiDP = pyart.retrieve.compute_spectral_differential_phase(
        psr,
        use_rhohv=use_rhohv,
        srhohv_field=srhohv_field,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(sPhiDP["standard_name"], sPhiDP)

    return new_dataset, ind_rad


def process_spectral_rhohv(procstatus, dscfg, radar_list=None):
    """
    Computes the spectral RhoHV

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            ("ShhADU" and "SvvADU") or ("ShhADUu" and "SvvADUu"), and,
            ("sNADUh" and "sNADUv")
        subtract_noise : Bool
            If True noise will be subtracted from the signal
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "sRhoHV" if "ShhADU" and "SvvADU" were provided
        "sRhoHVu" if "ShhADUu" and "SvvADUu" were provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "ShhADUu"):
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("SvvADU", "SvvADUu"):
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUv":
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_h_field not in psr.fields or signal_v_field not in psr.fields:
        warn("Unable to obtain spectral RhoHV. " + "Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)

    sRhoHV = pyart.retrieve.compute_spectral_rhohv(
        psr,
        subtract_noise=subtract_noise,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": deepcopy(psr)}
    new_dataset["radar_out"].fields = dict()
    new_dataset["radar_out"].add_field(sRhoHV["standard_name"], sRhoHV)

    return new_dataset, ind_rad


def process_pol_variables(procstatus, dscfg, radar_list=None):
    """
    Computes the polarimetric variables from the complex spectra

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            either a combination of signal and noise
            ("ShhADU" and "SvvADU") or ("ShhADUu" and "SvvADUu"), and,
            ("sNADUh" and "sNADUv"), or
            the power signal
            ("sPhhADU" and "sPvvADU") or ("sPhhADUu" and "sPvvADUu"), and
            ("sRhoHV" or "sRhoHVu")

        subtract_noise : Bool
            If True noise will be subtracted from the signal. Default False
        smooth_window : int or None
            Size of the moving Gaussian smoothing window. If none no smoothing
            will be applied. Default None
        variables : list of str
            list of variables to compute. Default dBZ
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the all outputs fields, that correspond to the
        specified "variables" keyword
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    pwr_h_field = None
    pwr_v_field = None
    srhohv_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "ShhADUu"):
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("SvvADU", "SvvADUu"):
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUv":
            noise_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPhhADU", "sPhhADUu"):
            pwr_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPvvADU", "sPvvADUu"):
            pwr_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sRhoHV", "sRhoHVu"):
            srhohv_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    use_pwr = False
    if pwr_h_field is not None or pwr_v_field is not None or srhohv_field is not None:
        use_pwr = True

    if not use_pwr and (
        signal_h_field not in psr.fields and signal_v_field not in psr.fields
    ):
        warn("Unable to obtain polarimetric variables. Missing fields")
        return None, None

    if use_pwr and (
        pwr_h_field not in psr.fields
        and pwr_h_field not in psr.fields
        and srhohv_field not in psr.fields
    ):
        warn("Unable to obtain polarimetric variables. Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)
    smooth_window = dscfg.get("smooth_window", None)
    variables = dscfg.get("variables", ["dBZ"])

    fields_list = []
    for variable in variables:
        fields_list.append(get_fieldname_pyart(variable))

    radar = pyart.retrieve.compute_pol_variables(
        psr,
        fields_list,
        use_pwr=use_pwr,
        subtract_noise=subtract_noise,
        smooth_window=smooth_window,
        srhohv_field=srhohv_field,
        pwr_h_field=pwr_h_field,
        pwr_v_field=pwr_v_field,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": radar}

    return new_dataset, ind_rad


def process_noise_power(procstatus, dscfg, radar_list=None):
    """
    Computes the noise power from the spectra

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "ShhADU" or "SvvADU" or "ShhADUu" or "SvvADUu"
        units : str
            The units of the returned signal. Can be 'ADU', 'dBADU' or 'dBm'
        navg : int
            Number of spectra averaged
        rmin : int
            Range from which the data is used to estimate the noise
        nnoise_min : int
            Minimum number of samples to consider the estimated noise power
            valid
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "NdBADUh" or "NdBADUv", or
        "NdBmh" or "NdBmv", or
        "Nh" or "Nv"
        depending on which input datatype and units were provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "SvvADU", "ShhADUu", "SvvADUu"):
            signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if signal_field not in psr.fields:
        warn("Unable to obtain spectral noise power. Missing field " + signal_field)
        return None, None

    units = dscfg.get("units", "ADU")
    navg = dscfg.get("navg", 1)
    rmin = dscfg.get("rmin", 0.0)
    nnoise_min = dscfg.get("nnoise_min", 100)

    noise = pyart.retrieve.compute_noise_power(
        psr,
        units=units,
        navg=navg,
        rmin=rmin,
        nnoise_min=nnoise_min,
        signal_field=signal_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(noise["standard_name"], noise)

    return new_dataset, ind_rad


def process_reflectivity(procstatus, dscfg, radar_list=None):
    """
    Computes reflectivity from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain
            "sdBZ" or sdBZv" or "sdBuZ" or "sdBuZv"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "dBZ", "dBZv",
        "dBuZ" or "dBuZv" depending on the provided input datatype
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sdBZ", "sdBZv", "sdBuZ", "sdBuZv"):
            sdBZ_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sdBZ_field not in psr.fields:
        warn("Unable to obtain reflectivity. " + "Missing field " + sdBZ_field)
        return None, None

    dBZ = pyart.retrieve.compute_reflectivity(psr, sdBZ_field=sdBZ_field)

    reflectivity_field = "reflectivity"
    if datatype in ("sdBZv", "sdBuZv"):
        reflectivity_field += "vv"

    if datatype in ("sdBuZ", "sdBuZv"):
        reflectivity_field = "unfiltered_" + reflectivity_field

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(reflectivity_field, dBZ)

    return new_dataset, ind_rad


def process_differential_reflectivity(procstatus, dscfg, radar_list=None):
    """
    Computes differential reflectivity from the horizontal and vertical
    spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "sdBZ" and "sdBZv", or
            "sdBuZ" and "sdBZuv"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output fields
        "ZDR" or "ZDRu" depending on the specified input datatype
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sdBZ", "sdBuZ"):
            sdBZ_field = get_fieldname_pyart(datatype)
        elif datatype in ("sdBZv", "sdBuZv"):
            sdBZv_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sdBZ_field not in psr.fields or sdBZv_field not in psr.fields:
        warn("Unable to obtain differential reflectivity. " + "Missing fields.")
        return None, None

    zdr = pyart.retrieve.compute_differential_reflectivity(
        psr, sdBZ_field=sdBZ_field, sdBZv_field=sdBZv_field
    )

    zdr_field = "differential_reflectivity"
    if "unfiltered" in sdBZ_field:
        zdr_field = "unfiltered_" + zdr_field

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(zdr_field, zdr)

    return new_dataset, ind_rad


def process_differential_phase(procstatus, dscfg, radar_list=None):
    """
    Computes the differential phase from the spectral differential phase and
    the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "uPhiDPu" or
        "uPhiDP" depending on the provided datatypes
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sdBZ", "sdBZv", "sdBuZ", "sdBuZv"):
            sdBZ_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPhiDP", "sPhiDPu"):
            sPhiDP_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sdBZ_field not in psr.fields or sPhiDP_field not in psr.fields:
        warn("Unable to obtain PhiDP. Missing fields")
        return None, None

    uphidp = pyart.retrieve.compute_differential_phase(
        psr, sdBZ_field=sdBZ_field, sPhiDP_field=sPhiDP_field
    )

    uphidp_field = "uncorrected_differential_phase"
    if "unfiltered" in sPhiDP_field:
        uphidp_field = "uncorrected_unfiltered_differential_phase"

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(uphidp_field, uphidp)

    return new_dataset, ind_rad


def process_rhohv(procstatus, dscfg, radar_list=None):
    """
    Computes RhoHV from the complex spectras

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            either a combination of signal and noise
            ("ShhADU" and "SvvADU") or ("ShhADUu" and "SvvADUu"), and,
            ("sNADUh" and "sNADUv"), or
            the power signal
            ("sPhhADU" and "sPvvADU") or ("sPhhADUu" and "sPvvADUu"), and
            optionally ("sRhoHV" or "sRhoHVu")

        subtract_noise : Bool
            If True noise will be subtracted from the signal
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field "RhoHV"
        or "uRhoHV" depending on the provided datatypes

    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    pwr_h_field = None
    pwr_v_field = None
    srhohv_field = None
    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("ShhADU", "ShhADUu"):
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("SvvADU", "SvvADUu"):
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUh":
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == "sNADUv":
            noise_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPhhADU", "sPhhADUu"):
            pwr_h_field = get_fieldname_pyart(datatype)
        elif datatype in ("sPvvADU", "sPvvADUu"):
            pwr_v_field = get_fieldname_pyart(datatype)
        elif datatype in ("sRhoHV", "sRhoHVu"):
            srhohv_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if srhohv_field is not None:
        use_rhohv = True
        rhohv_field = "cross_correlation_ratio"
        if "unfiltered" in srhohv_field:
            rhohv_field = "unfiltered_cross_correlation_ratio"
    else:
        use_rhohv = False
        rhohv_field = "cross_correlation_ratio"
        if "unfiltered" in signal_h_field:
            rhohv_field = "unfiltered_cross_correlation_ratio"

    if not use_rhohv and (
        signal_h_field not in psr.fields or signal_v_field not in psr.fields
    ):
        warn("Unable to obtain RhoHV. Missing fields")
        return None, None

    if use_rhohv and (
        srhohv_field not in psr.fields
        or pwr_h_field not in psr.fields
        or pwr_v_field not in psr.fields
    ):
        warn("Unable to obtain RhoHV. Missing fields")
        return None, None

    subtract_noise = dscfg.get("subtract_noise", False)

    rhohv = pyart.retrieve.compute_rhohv(
        psr,
        use_rhohv=use_rhohv,
        subtract_noise=subtract_noise,
        srhohv_field=srhohv_field,
        pwr_h_field=pwr_h_field,
        pwr_v_field=pwr_v_field,
        signal_h_field=signal_h_field,
        signal_v_field=signal_v_field,
        noise_h_field=noise_h_field,
        noise_v_field=noise_v_field,
    )

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(rhohv_field, rhohv)

    return new_dataset, ind_rad


def process_Doppler_velocity(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler velocity from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "sdBZ" or "sdBZv" or "sdBuZ" or "sdBuZv"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "V" if "sdBZ" was provided
        "Vv" if "sdBZv" was provided
        "Vu" if "sdBuZ" was provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sdBZ", "sdBZv", "sdBuZ", "sdBuZv"):
            sdBZ_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sdBZ_field not in psr.fields:
        warn("Unable to obtain Doppler velocity. " + "Missing field " + sdBZ_field)
        return None, None

    vel = pyart.retrieve.compute_Doppler_velocity(psr, sdBZ_field=sdBZ_field)

    vel_field = "velocity"
    if datatype in ("sdBZv", "sdBuZv"):
        vel_field += "_vv"
    if datatype in ("sdBuZ", "sdBuZv"):
        vel_field = "unfiltered_" + vel_field

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(vel_field, vel)

    return new_dataset, ind_rad


def process_Doppler_width(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectrum width from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types, must contain,
            "sdBZ" or "sdBZv" or "sdBuZ" or "sdBuZv"
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output field
        "W" if "sdBZ" was provided
        "Wv" if "sdBZv" was provided
        "Wu" if "sdBuZ" was provided
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ("sdBZ", "sdBZv", "sdBuZ", "sdBuZv"):
            sdBZ_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8]) - 1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    psr = radar_list[ind_rad]

    if sdBZ_field not in psr.fields:
        warn(
            "Unable to obtain Doppler spectrum width. " + "Missing field " + sdBZ_field
        )
        return None, None

    width = pyart.retrieve.compute_Doppler_width(psr, sdBZ_field=sdBZ_field)

    width_field = "spectrum_width"
    if datatype in ("sdBZv", "sdBuZv"):
        width_field += "_vv"
    if datatype in ("sdBuZ", "sdBuZv"):
        width_field = "unfiltered_" + width_field

    # prepare for exit
    new_dataset = {"radar_out": pyart.util.radar_from_spectra(psr)}
    new_dataset["radar_out"].add_field(width_field, width)

    return new_dataset, ind_rad



#!/usr/bin/env python
""" Python Radar Toolkit

Pyrad is a Python module containing
the utilities that run the MeteoSwiss radar processing framework.
It is designed so that it accepts a growing number of radar data types.
The core of the processing is performed by the module Py-ART.

"""

import os
import subprocess
import glob
import builtins
from datetime import datetime
import getpass
from setuptools import setup, find_packages

DOCLINES = __doc__.split("\n")

CLASSIFIERS = """\
    Development Status :: 5 - Production/Stable
    Intended Audience :: Science/Research
    Intended Audience :: Developers
    License :: OSI Approved :: BSD License
    Programming Language :: Python
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3.6
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: 3.8
    Programming Language :: Python :: 3.9
    Programming Language :: Python :: 3.10
    Programming Language :: C
    Programming Language :: Cython
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Atmospheric Science
    Operating System :: POSIX :: Linux
    Operating System :: MacOS :: MacOS X
    Operating System :: Microsoft :: Windows
    Framework :: Matplotlib
"""


NAME = "pyrad_mch"
MAINTAINER = "MeteoSwiss Pyrad Developers"
MAINTAINER_EMAIL = "daniel.wolfensberger@meteoswiss.ch"
DESCRIPTION = DOCLINES[0]
LONG_DESCRIPTION = "\n".join(DOCLINES[2:])
URL = "https://github.com/MeteoSwiss/pyrad.git"
DOWNLOAD_URL = "https://github.com/MeteoSwiss/pyrad.git"
LICENSE = "BSD"
CLASSIFIERS = list(filter(None, CLASSIFIERS.split("\n")))
PLATFORMS = ["Linux"]
MAJOR = 2
MINOR = 0
MICRO = 3
ISRELEASED = True
VERSION = "%d.%d.%d" % (MAJOR, MINOR, MICRO)
SCRIPTS = glob.glob("scripts/*")
COMPILE_DATE_TIME = datetime.utcnow().strftime("%Y-%m-%d %H:%M")
USERNAME = getpass.getuser()


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists("MANIFEST"):
    os.remove("MANIFEST")

# This is a bit hackish: we are setting a global variable so that the main
# pyrad __init__ can detect if it is being loaded by the setup routine, to
# avoid attempting to load components that aren't built yet. While ugly, it's
# a lot more robust than what was previously being used.
builtins.__PYRAD_SETUP__ = True


def write_version_py(filename="pyrad/version.py"):
    cnt = """
# THIS FILE IS GENERATED FROM PYRAD_PROC SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
compile_date_time = '%(compile_date_time)s'
username = '%(username)s'
release = %(isrelease)s

if not release:
    version = full_version
"""
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of pyrad.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists("../../.git"):
        GIT_REVISION = git_version()
    elif os.path.exists("pyrad/version.py"):
        # must be a source distribution, use existing version file
        try:
            from pyrad.version import git_revision as GIT_REVISION
        except ImportError:
            raise ImportError(
                "Unable to import git_revision. Try removing "
                "pyrad/version.py and the build directory "
                "before building."
            )
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += ".dev+" + GIT_REVISION[:7]

    a = open(filename, "w")
    try:
        a.write(
            cnt
            % {
                "version": VERSION,
                "full_version": FULLVERSION,
                "git_revision": GIT_REVISION,
                "compile_date_time": COMPILE_DATE_TIME,
                "username": USERNAME,
                "isrelease": str(ISRELEASED),
            }
        )
    finally:
        a.close()


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True,
    )

    config.add_subpackage("pyrad")
    config.add_data_files(("pyrad", "*.txt"))

    return config


if __name__ == "__main__":
    write_version_py()
    setup(
        name=NAME,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        url=URL,
        packages=find_packages(),
        include_package_data=True,
        configuration=configuration,
        scripts=SCRIPTS,
        license=LICENSE,
        platforms=PLATFORMS,
        classifiers=CLASSIFIERS,
        version=VERSION,
        zip_safe=False,
    )
